@@ -379,7 +379,8 @@ export abstract class BaseSettingsPage {
       localSettings = {
         serverUrl: registrySettings.serverUrl || this.currentSettings.serverUrl,
         username: registrySettings.username || this.currentSettings.username,
-        usernameFromJupyterHub: registrySettings.usernameFromJupyterHub || this.currentSettings.usernameFromJupyterHub
+        usernameFromJupyterHub: registrySettings.usernameFromJupyterHub || this.currentSettings.usernameFromJupyterHub,
+        bonnieUrl: registrySettings.bonnieUrl || this.currentSettings.bonnieUrl
       };
       console.log('🔧 SETTINGS: Local settings loaded from registry');
     } catch (error) {
@@ -450,6 +451,7 @@ export abstract class BaseSettingsPage {
       serverUrl: localSettings.serverUrl!,
       username: localSettings.username!,
       usernameFromJupyterHub: localSettings.usernameFromJupyterHub!,
+      bonnieUrl: localSettings.bonnieUrl, // Include bonnieUrl from local settings
 
       // Remote settings (from server)
       maxMessages: remoteSettings.maxMessages!,
@@ -549,128 +551,55 @@ export abstract class BaseSettingsPage {
     // Update the global settings object immediately for message handler access
     (window as any).escobarCurrentSettings = newSettings;
 
-    // Save settings using the new server-side mechanism (NO reconnection)
-    this.saveSettingsToServerAndBrowserOnly(newSettings);
+    // Save settings to server only (no localStorage fallbacks)
+    this.saveSettingsToServerOnly(newSettings);
 
     return true;
   }
 
   /**
-   * Save settings to server and browser storage
+   * Save settings to server only (clean, no fallbacks)
    * @param settings The settings to save
    */
-  private async saveSettingsToServerAndBrowser(settings: IChatSettings): Promise<void> {
-    console.log('🔧 SETTINGS: Starting settings save process...');
-
-    // Set flag to prevent reconnection during save
-    (window as any).escobarSettingsSaveInProgress = true;
+  private async saveSettingsToServerOnly(settings: IChatSettings): Promise<void> {
+    console.log('🔧 SETTINGS: Saving remote settings to server only');
 
     try {
       // Get the message handler from the global window object
       const messageHandler = (window as any).escobarMessageHandler as MessageHandler;
 
-      if (messageHandler) {
-        console.log('🔧 SETTINGS: Saving remote settings to server');
-        // Save most settings to server (excluding serverUrl and username)
-        await messageHandler.saveSettingsToServer({
-          maxMessages: settings.maxMessages,
-          voittaApiKey: settings.voittaApiKey,
-          openaiApiKey: settings.openaiApiKey,
-          anthropicApiKey: settings.anthropicApiKey,
-          geminiApiKey: settings.geminiApiKey,
-          proxyPort: settings.proxyPort,
-          primaryModel: settings.primaryModel,
-          secondaryProvider: settings.secondaryProvider,
-          imageParseProvider: settings.imageParseProvider
-        });
-        console.log('🔧 SETTINGS: Remote settings saved to server successfully');
-      } else {
-        console.warn('🔧 SETTINGS: MessageHandler not available, skipping server save');
+      if (!messageHandler) {
+        throw new Error('Message handler not available - cannot save settings to server');
       }
+
+      console.log('📤 PROTOCOL: Sending saveSettings request to server');
+      // Save remote settings to server only (no localStorage fallback)
+      await messageHandler.saveSettingsToServer({
+        maxMessages: settings.maxMessages,
+        voittaApiKey: settings.voittaApiKey,
+        openaiApiKey: settings.openaiApiKey,
+        anthropicApiKey: settings.anthropicApiKey,
+        geminiApiKey: settings.geminiApiKey,
+        proxyPort: settings.proxyPort,
+        primaryModel: settings.primaryModel,
+        secondaryProvider: settings.secondaryProvider,
+        imageParseProvider: settings.imageParseProvider
+      });
+
+      console.log('✅ SETTINGS: Remote settings saved to server successfully');
+
+      // Hide settings page after successful save
+      this.hide();
+
     } catch (error) {
-      console.error('🔧 SETTINGS: Failed to save remote settings to server:', error);
-      // Continue with local storage even if server save fails
+      console.error('❌ SETTINGS: Failed to save settings to server:', error);
+
+      // Show clear error message to user
+      const errorMessage = error instanceof Error ? error.message : 'Unknown error occurred';
+      alert(`Failed to save settings to server:\n\n${errorMessage}\n\nPlease check your connection and try again.`);
+
+      // Don't hide the settings page so user can retry
     }
-
-    // Save ONLY local settings to browser storage (JupyterLab settings registry)
-    // Remote settings (API keys, models) are saved to server only
-    try {
-      const settingsPlugin = await this.settingsRegistry.load('escobar:plugin');
-
-      console.log('📋 SCHEMA: Saving only local settings to registry');
-      // Only save the 3 fields allowed by our schema
-      await settingsPlugin.set('serverUrl', settings.serverUrl);
-      await settingsPlugin.set('username', settings.username);
-      await settingsPlugin.set('usernameFromJupyterHub', settings.usernameFromJupyterHub);
-
-      console.log('📋 SCHEMA: Registry save successful');
-    } catch (error) {
-      console.error('📋 SCHEMA: Registry save failed:', error);
-    }
-
-    // Also save to localStorage as fallback
-    try {
-      localStorage.setItem('escobar-settings', JSON.stringify(settings));
-      console.log('🔧 SETTINGS: Settings saved to localStorage as fallback');
-    } catch (error) {
-      console.error('🔧 SETTINGS: Failed to save settings to localStorage:', error);
-    }
-
-    // Clear the flag after save is complete
-    (window as any).escobarSettingsSaveInProgress = false;
-
-    console.log('🔧 SETTINGS: Settings save process completed');
-
-    // Hide settings page after successful save
-    this.hide();
-  }
-
-  /**
-   * Save settings to server and localStorage only (no registry save, no reconnection)
-   * @param settings The settings to save
-   */
-  private async saveSettingsToServerAndBrowserOnly(settings: IChatSettings): Promise<void> {
-    console.log('🔧 SETTINGS: Starting user settings save process (no reconnection)...');
-
-    try {
-      // Get the message handler from the global window object
-      const messageHandler = (window as any).escobarMessageHandler as MessageHandler;
-
-      if (messageHandler) {
-        console.log('🔧 SETTINGS: Saving remote settings to server');
-        // Save most settings to server (excluding serverUrl and username)
-        await messageHandler.saveSettingsToServer({
-          maxMessages: settings.maxMessages,
-          voittaApiKey: settings.voittaApiKey,
-          openaiApiKey: settings.openaiApiKey,
-          anthropicApiKey: settings.anthropicApiKey,
-          geminiApiKey: settings.geminiApiKey,
-          proxyPort: settings.proxyPort,
-          primaryModel: settings.primaryModel,
-          secondaryProvider: settings.secondaryProvider,
-          imageParseProvider: settings.imageParseProvider
-        });
-        console.log('🔧 SETTINGS: Remote settings saved to server successfully');
-      } else {
-        console.warn('🔧 SETTINGS: MessageHandler not available, skipping server save');
-      }
-    } catch (error) {
-      console.error('🔧 SETTINGS: Failed to save remote settings to server:', error);
-      // Continue with local storage even if server save fails
-    }
-
-    // Save to localStorage as fallback (no registry save to avoid triggering reconnection)
-    try {
-      localStorage.setItem('escobar-settings', JSON.stringify(settings));
-      console.log('🔧 SETTINGS: Settings saved to localStorage as fallback');
-    } catch (error) {
-      console.error('🔧 SETTINGS: Failed to save settings to localStorage:', error);
-    }
-
-    console.log('🔧 SETTINGS: User settings save process completed (no reconnection)');
-
-    // Hide settings page after successful save
-    this.hide();
   }
 }
 
@@ -871,11 +800,6 @@ export class JupyterHubSettingsPage extends BaseSettingsPage {
     const usernameInput = document.getElementById('escobar-username') as HTMLInputElement;
     const proxyPortInput = document.getElementById('escobar-proxy-port') as HTMLInputElement;
 
-    // Get model dropdown elements
-    const primaryModelSelect = document.getElementById('escobar-primary-model') as HTMLSelectElement;
-    const secondaryProviderSelect = document.getElementById('escobar-secondary-provider') as HTMLSelectElement;
-    const imageParseProviderSelect = document.getElementById('escobar-image-parse-provider') as HTMLSelectElement;
-
     // Update values with current settings
     if (maxMessagesInput) maxMessagesInput.value = this.currentSettings.maxMessages.toString();
     if (serverUrlInput) serverUrlInput.value = this.currentSettings.serverUrl;
@@ -887,9 +811,44 @@ export class JupyterHubSettingsPage extends BaseSettingsPage {
     if (usernameInput) usernameInput.value = this.currentSettings.username;
     if (proxyPortInput) proxyPortInput.value = (this.currentSettings.proxyPort || 3000).toString();
 
-    // Note: Model dropdown selections are now handled automatically by the two-phase approach
-    // Phase 1: Dropdowns are created with saved values immediately
-    // Phase 2: When models load, dropdowns are re-populated while preserving saved values
+    // Create model dropdowns now that settings are loaded
+    // Get placeholder containers
+    const primaryModelContainer = document.getElementById('primary-model-container');
+    const secondaryProviderContainer = document.getElementById('secondary-provider-container');
+    const imageParseProviderContainer = document.getElementById('image-parse-provider-container');
+
+    if (primaryModelContainer) {
+      primaryModelContainer.innerHTML = ''; // Clear placeholder
+      const primaryModelGroup = this.createModelDropdown(
+        'escobar-primary-model',
+        'Primary Model',
+        'The main AI model used for chat responses.',
+        this.currentSettings.primaryModel
+      );
+      primaryModelContainer.appendChild(primaryModelGroup);
+    }
+
+    if (secondaryProviderContainer) {
+      secondaryProviderContainer.innerHTML = ''; // Clear placeholder
+      const secondaryProviderGroup = this.createModelDropdown(
+        'escobar-secondary-provider',
+        'Secondary Provider',
+        'Secondary AI model for specialized tasks.',
+        this.currentSettings.secondaryProvider
+      );
+      secondaryProviderContainer.appendChild(secondaryProviderGroup);
+    }
+
+    if (imageParseProviderContainer) {
+      imageParseProviderContainer.innerHTML = ''; // Clear placeholder
+      const imageParseProviderGroup = this.createModelDropdown(
+        'escobar-image-parse-provider',
+        'Image Parse Provider',
+        'AI model used for image analysis and parsing.',
+        this.currentSettings.imageParseProvider
+      );
+      imageParseProviderContainer.appendChild(imageParseProviderGroup);
+    }
   }
 
   /**
@@ -1002,6 +961,33 @@ export class ConnectionSettingsPage extends BaseSettingsPage {
     usernameGroup.appendChild(usernameInput);
     form.appendChild(usernameGroup);
 
+    // Bonnie URL field
+    const bonnieUrlGroup = this.createFormGroup(
+      'escobar-connection-bonnie-url',
+      'Bonnie URL',
+      'The WebSocket URL for the Bonnie backend server. If specified, this overrides the WEBSOCKET_PROXY_TARGET environment variable. (Optional)'
+    );
+
+    const bonnieUrlInput = document.createElement('input');
+    bonnieUrlInput.id = 'escobar-connection-bonnie-url';
+    bonnieUrlInput.className = 'escobar-settings-input';
+    bonnieUrlInput.type = 'text';
+    bonnieUrlInput.placeholder = 'ws://bonnie:8777/ws';
+    bonnieUrlInput.value = this.currentSettings.bonnieUrl || '';
+    bonnieUrlGroup.appendChild(bonnieUrlInput);
+
+    // Add help text for Bonnie URL
+    const bonnieUrlHelp = document.createElement('div');
+    bonnieUrlHelp.className = 'escobar-settings-note';
+    bonnieUrlHelp.style.fontSize = '0.85em';
+    bonnieUrlHelp.style.fontStyle = 'italic';
+    bonnieUrlHelp.style.marginTop = '5px';
+    bonnieUrlHelp.style.color = '#666';
+    bonnieUrlHelp.innerHTML = 'Examples: <code>ws://localhost:8777/ws</code>, <code>wss://api.example.com/ws</code>';
+    bonnieUrlGroup.appendChild(bonnieUrlHelp);
+
+    form.appendChild(bonnieUrlGroup);
+
     // Warning message
     const warningMessage = document.createElement('div');
     warningMessage.style.cssText = `
@@ -1035,9 +1021,11 @@ export class ConnectionSettingsPage extends BaseSettingsPage {
   protected updateFormFields(): void {
     const serverUrlInput = document.getElementById('escobar-connection-server-url') as HTMLInputElement;
     const usernameInput = document.getElementById('escobar-connection-username') as HTMLInputElement;
+    const bonnieUrlInput = document.getElementById('escobar-connection-bonnie-url') as HTMLInputElement;
 
     if (serverUrlInput) serverUrlInput.value = this.currentSettings.serverUrl;
     if (usernameInput) usernameInput.value = this.currentSettings.username;
+    if (bonnieUrlInput) bonnieUrlInput.value = this.currentSettings.bonnieUrl || '';
   }
 
   /**
@@ -1046,10 +1034,12 @@ export class ConnectionSettingsPage extends BaseSettingsPage {
   protected saveSettings(): void {
     const serverUrlInput = document.getElementById('escobar-connection-server-url') as HTMLInputElement;
     const usernameInput = document.getElementById('escobar-connection-username') as HTMLInputElement;
+    const bonnieUrlInput = document.getElementById('escobar-connection-bonnie-url') as HTMLInputElement;
 
     // Validate inputs
     const serverUrl = serverUrlInput.value.trim();
     const username = usernameInput.value.trim();
+    const bonnieUrl = bonnieUrlInput.value.trim();
 
     if (!serverUrl) {
       alert('Server URL is required');
@@ -1061,11 +1051,26 @@ export class ConnectionSettingsPage extends BaseSettingsPage {
       return;
     }
 
+    // Validate Bonnie URL format if provided (allow empty string to clear the setting)
+    if (bonnieUrl && bonnieUrl.length > 0) {
+      try {
+        const url = new URL(bonnieUrl);
+        if (!['ws:', 'wss:'].includes(url.protocol)) {
+          alert('Bonnie URL must use ws:// or wss:// protocol');
+          return;
+        }
+      } catch (error) {
+        alert('Invalid Bonnie URL format. Please enter a valid WebSocket URL.');
+        return;
+      }
+    }
+
     // Create new settings object with only connection changes
     const newSettings: IChatSettings = {
       ...this.currentSettings,
       serverUrl: serverUrl,
-      username: username
+      username: username,
+      bonnieUrl: bonnieUrl || undefined
     };
 
     console.log('🔗 CONNECTION: Saving connection settings');
@@ -1086,6 +1091,8 @@ export class ConnectionSettingsPage extends BaseSettingsPage {
       await settingsPlugin.set('serverUrl', settings.serverUrl);
       await settingsPlugin.set('username', settings.username);
       await settingsPlugin.set('usernameFromJupyterHub', settings.usernameFromJupyterHub);
+      // Always save bonnieUrl, even if it's empty (to clear the setting)
+      await settingsPlugin.set('bonnieUrl', settings.bonnieUrl || '');
 
       console.log('🔗 CONNECTION: Registry save successful');
 
@@ -1098,14 +1105,6 @@ export class ConnectionSettingsPage extends BaseSettingsPage {
       // Call onSave callback (this will trigger reconnection)
       this.onSave(settings);
 
-      // Also save to localStorage as fallback
-      try {
-        localStorage.setItem('escobar-settings', JSON.stringify(settings));
-        console.log('🔗 CONNECTION: Settings saved to localStorage as fallback');
-      } catch (error) {
-        console.error('🔗 CONNECTION: Failed to save to localStorage:', error);
-      }
-
       console.log('🔗 CONNECTION: Connection settings save completed');
 
       // Hide settings page
@@ -1113,7 +1112,7 @@ export class ConnectionSettingsPage extends BaseSettingsPage {
 
     } catch (error) {
       console.error('🔗 CONNECTION: Failed to save connection settings:', error);
-      alert('Failed to save connection settings. Please try again.');
+      alert(`Failed to save connection settings:\n\n${error instanceof Error ? error.message : 'Unknown error'}\n\nPlease try again.`);
     }
   }
 }
